import { Injectable, Inject } from '@angular/core';
import {
  HttpClient,
  HttpEventType,
  HttpResponse,
  HttpRequest,
  HttpProgressEvent
} from '@angular/common/http';

import { Observable } from 'rxjs';
import { map } from 'rxjs/operators';

import { _MODULE_CONFIG } from '../types/token';
import { ModuleConfig } from '../types/module-config';
import {
  FileProfileSaveRequest,
  FileProfileSaveResponse,
  encodeFileProfileSave,
  decodeFileProfileSave
} from '../apis/file-profile-save';
import {
  FileTalkDownloadRequest,
  encodeFileTalkDownload
} from '../apis/file-talk-download';
import {
  FileTalkSaveRequest,
  FileTalkSaveResponse,
  encodeFileTalkSave,
  decodeFileTalkSave
} from '../apis/file-talk-save';
import {
  FileTalkShareRequest,
  FileTalkShareResponse,
  encodeFileTalkShare,
  decodeFileTalkShare
} from '../apis/file-talk-share';
import {
  MassTalkDownloadRequest,
  MassTalkDownloadResponse,
  encodeMassTalkDownload,
  decodeMassTalkDownload
} from '../apis/mass-talk-download';
import {
  MassTalkSaveRequest,
  MassTalkSaveResponse,
  encodeMassTalkSave,
  decodeMassTalkSave
} from '../apis/mass-talk-save';
import {
  TransMassTalkDownloadRequest,
  TransMassTalkDownloadResponse,
  encodeTransMassTalkDownload,
  decodeTransMassTalkDownload
} from '../apis/trans-mass-talk-download';
import {
  TransMassTalkSaveRequest,
  TransMassTalkSaveResponse,
  encodeTransMassTalkSave,
  decodeTransMassTalkSave
} from '../apis/trans-mass-talk-save';
import {
  TranslationReqRequest,
  TranslationReqResponse,
  encodeTranslationReq,
  decodeTranslationReq
} from '../apis/translation-req';
import {
  TranslationSaveRequest,
  TranslationSaveResponse,
  encodeTranslationSave,
  decodeTranslationSave
} from '../apis/translation-save';

@Injectable({
  providedIn: 'root'
})
export class CommonApiService {
  constructor(
    @Inject(_MODULE_CONFIG) private moduleConfig: ModuleConfig,
    private httpClient: HttpClient
  ) {}

  public fileProfileSave(
    req: FileProfileSaveRequest,
    fileProfileSaveUrl?: string
  ): Observable<FileProfileSaveResponse> {
    return this.httpClient
      .post<any>(
        !!fileProfileSaveUrl
          ? fileProfileSaveUrl
          : this.moduleConfig.urls.fileProfileSave,
        {},
        {
          params: encodeFileProfileSave(req)
        }
      )
      .pipe(map(res => decodeFileProfileSave(res)));
  }

  public fileTalkDownload(
    req: FileTalkDownloadRequest,
    fileTalkDownloadUrl?: string
  ): Observable<Blob> {
    return this.httpClient.post<Blob>(
      !!fileTalkDownloadUrl
        ? fileTalkDownloadUrl
        : this.moduleConfig.urls.fileTalkDownload,
      { responseType: 'blob' },
      {
        params: encodeFileTalkDownload(req)
      }
    );
  }

  public fileTalkSave(
    req: FileTalkSaveRequest,
    fileTalkSaveUrl?: string
  ): Observable<FileTalkSaveResponse | HttpProgressEvent> {
    const httpReq = new HttpRequest(
      'POST',
      !!fileTalkSaveUrl ? fileTalkSaveUrl : this.moduleConfig.urls.fileTalkSave,
      encodeFileTalkSave(req),
      { reportProgress: true }
    );

    return this.httpClient.request(httpReq).pipe(
      map(event => {
        if (event instanceof HttpResponse) {
          return decodeFileTalkSave(event);
        } else if (HttpEventType.UploadProgress === event.type) {
          return event;
        }
      })
    );
  }

  public fileTalkShare(
    req: FileTalkShareRequest
  ): Observable<FileTalkShareResponse> {
    return this.httpClient
      .post<any>(
        this.moduleConfig.urls.fileTalkShare,
        {},
        {
          params: encodeFileTalkShare(req)
        }
      )
      .pipe(map(res => decodeFileTalkShare(res)));
  }

  public massTalkDownload(
    req: MassTalkDownloadRequest
  ): Observable<MassTalkDownloadResponse> {
    return this.httpClient
      .post<any>(
        this.moduleConfig.urls.massTalkDownload,
        {},
        {
          params: encodeMassTalkDownload(req),
          responseType: 'text' as 'json'
        }
      )
      .pipe(map(res => decodeMassTalkDownload(res)));
  }

  public massTalkSave(
    req: MassTalkSaveRequest
  ): Observable<MassTalkSaveResponse> {
    return this.httpClient
      .post<any>(
        this.moduleConfig.urls.massTalkSave,
        {},
        {
          params: encodeMassTalkSave(req),
          responseType: 'text' as 'json'
        }
      )
      .pipe(map(res => decodeMassTalkSave(res)));
  }

  public transMassTalkDownload(
    req: TransMassTalkDownloadRequest
  ): Observable<TransMassTalkDownloadResponse> {
    return this.httpClient
      .post<any>(
        this.moduleConfig.urls.transMassTalkDownload,
        {},
        {
          params: encodeTransMassTalkDownload(req)
        }
      )
      .pipe(map(res => decodeTransMassTalkDownload(res)));
  }

  public transMassTalkSave(
    req: TransMassTalkSaveRequest
  ): Observable<TransMassTalkSaveResponse> {
    return this.httpClient
      .post<any>(
        this.moduleConfig.urls.transMassTalkSave,
        {},
        {
          params: encodeTransMassTalkSave(req)
        }
      )
      .pipe(map(res => decodeTransMassTalkSave(res)));
  }

  public translationReq(
    req: TranslationReqRequest
  ): Observable<TranslationReqResponse> {
    return this.httpClient
      .post<any>(
        this.moduleConfig.urls.translationReq,
        {},
        {
          params: encodeTranslationReq(req)
        }
      )
      .pipe(map(res => decodeTranslationReq(res)));
  }

  public translationSave(
    req: TranslationSaveRequest
  ): Observable<TranslationSaveResponse> {
    return this.httpClient
      .post<any>(
        this.moduleConfig.urls.translationSave,
        {},
        {
          params: encodeTranslationSave(req)
        }
      )
      .pipe(map(res => decodeTranslationSave(res)));
  }
}
